"""Artistools - spectra related functions."""

__all__ = ["plot"]

from artistools.nonthermal import solvespencerfanocmd as solvespencerfanocmd
from artistools.nonthermal.plotnonthermal import addargs as addargs
from artistools.nonthermal.plotnonthermal import main as plot
